from ...import_utils import unavailable_class
from ...version_utils import has_minimal_version

if has_minimal_version("blissdata", "1"):
    try:
        import bliss  # noqa F401
    except ImportError as ex:
        PersistentNdArray = unavailable_class(ex)
    else:
        from .ndarrayv1 import PersistentNdArray
else:
    from .ndarrayv0 import PersistentNdArray  # noqa F401
