from .. import testing
from ..processors.streamline import DemoStreamlineScanner
from .utils import streamline


def streamline_demo():
    test_streamline_without_autocalibration()
    test_streamline_with_autocalibration()


@testing.integration_fixture
def streamline_scanner():
    return DemoStreamlineScanner()


@testing.integration_test
def test_streamline_with_autocalibration(streamline_scanner):
    streamline.test_streamline_with_autocalibration(streamline_scanner)


@testing.integration_test
def test_streamline_without_autocalibration(streamline_scanner):
    streamline.test_streamline_without_autocalibration(streamline_scanner)
