"""Bliss-side for Flint BM32 XRR plots."""

from ..flint import BasePlot


class Bm32XrrAnalysisPlot(BasePlot):
    WIDGET = "blissoda.bm32.xrr_widgets.Bm32XrrAnalysisWidget"

    def clear_data(self) -> None:
        self.submit("clear_data")

    def set_result_data(self, result: dict, sample_name: str = "") -> None:
        self.submit("set_result_data", result, sample_name)
