from .. import testing
from ..processors.streamline_id31 import DemoId31StreamlineScanner
from .utils import streamline


def id31_streamline_demo():
    test_streamline_without_autocalibration()
    test_streamline_with_autocalibration()


def id31_streamline_auto_stop_demo():
    test_streamline_id31_auto_stop()


@testing.integration_fixture
def streamline_id31_scanner():
    return DemoId31StreamlineScanner()


@testing.integration_fixture
def streamline_id31_auto_stop_scanner():
    return DemoId31StreamlineScanner()


@testing.integration_test
def test_streamline_with_autocalibration(streamline_id31_scanner):
    streamline.test_streamline_with_autocalibration(streamline_id31_scanner)


@testing.integration_test
def test_streamline_without_autocalibration(streamline_id31_scanner):
    streamline.test_streamline_without_autocalibration(streamline_id31_scanner)


@testing.integration_test
def test_streamline_id31_auto_stop(streamline_id31_auto_stop_scanner):
    streamline.test_streamline_id31_auto_stop(streamline_id31_auto_stop_scanner)
