"""Bliss-side for Flint ID26 plots"""

from ..flint import ImageView
from ..flint import Plot1D


class AggregatedScansPlot(Plot1D):
    """Plot for aggregated scans."""


class RixsPlot(ImageView):
    """Plot for RIXS maps."""
