from .. import testing
from ..processors.id22 import DemoStScanProcessor


@testing.integration_fixture
def stscan_processor():
    yield DemoStScanProcessor()


@testing.integration_test
def id22_stscan_demo(stscan_processor):
    stscan_processor.submit_workflows()
